/* ======================================================================== *\
!
! *
! * This file is part of MARS, the MAGIC Analysis and Reconstruction
! * Software. It is distributed to you in the hope that it can be a useful
! * and timesaving tool in analysing Data of imaging Cerenkov telescopes.
! * It is distributed WITHOUT ANY WARRANTY.
! *
! * Permission to use, copy, modify and distribute this software and its
! * documentation for any purpose is hereby granted without fee,
! * provided that the above copyright notice appear in all copies and
! * that both that copyright notice and this permission notice appear
! * in supporting documentation. It is provided "as is" without express
! * or implied warranty.
! *
!
!
!   Author(s): Thomas Bretz  12/2000 <mailto:tbretz@uni-sw.gwdg.de>
!   Author(s): Harald Kornmayer 1/2001
!
!   Copyright: MAGIC Software Development, 2000-2001
!
!   Modified 4/7/2002 Abelardo Moralejo: now the dimension of fTrigger is
!    set dinamically, to allow an arbitrary large number of trigger 
!    conditions to be processed.
!
!
\* ======================================================================== */

#include "MMcTriggerRateCalc.h"

#include "MLog.h"
#include "MLogManip.h"

#include "MParList.h"
#include "MHMcRate.h"

#include "MMcEvt.hxx"
#include "MMcTrig.hxx"
#include "MMcRunHeader.hxx"
#include "MMcCorsikaRunHeader.h"

ClassImp(MMcTriggerRateCalc);

void MMcTriggerRateCalc::Init(int dim, float *trigbg, float simbg,
                              const char *name, const char *title)
{
    fName  = name  ? name  : "MMcTriggerRateCalc";
    fTitle = title ? title : "Task to calc the trigger rate ";

    fMcTrig = NULL;
    fMcRate = NULL;

    fTrigNSB = trigbg;
    fSimNSB = simbg;

    fPartId = -1; 

    fShowers = 0;
    fAnalShow = 0;

    fFirst = dim>0 ?   1 : -dim;
    fLast  = dim>0 ? dim : -dim;

    fNum = fLast-fFirst+1;
    fTrigger = new float[fNum];

    for (UInt_t i=0;i<fNum;i++)
      fTrigger[i]=0;

    AddToBranchList("MMcEvt.fPartId");
    AddToBranchList("MMcEvt.fImpact");
    AddToBranchList("MMcEvt.fEnergy");
    AddToBranchList("MMcEvt.fPhi");
    AddToBranchList("MMcEvt.fTheta");
    AddToBranchList("MMcEvt.fPhotElfromShower");
    AddToBranchList("MMcTrig", "fNumFirstLevel", fFirst, fLast);

}

// ReInit: read run header to get some info later:

Bool_t MMcTriggerRateCalc::ReInit(MParList *pList)
{
    fMcRunHeader = (MMcRunHeader*) pList->FindObject("MMcRunHeader");
    if (!fMcRunHeader)
    {
        *fLog << err << dbginf << "Error - MMcRunHeader not found... exit." << endl;
        return kFALSE;
    }

    fMcCorRunHeader = (MMcCorsikaRunHeader*) pList->FindObject("MMcCorsikaRunHeader");
    if (!fMcCorRunHeader)
    {
        *fLog << err << dbginf << "Error - MMcCorsikaRunHeader not found... exit." << endl;
        return kFALSE;
    }

    fShowers += fMcRunHeader->GetNumSimulatedShowers();

    if (fMcRunHeader->GetAllEvtsTriggered())
      {
	*fLog << endl << all << endl <<
	  "WARNING: the input data file contains only the" << endl << 
	  "events that triggered. I will assume the standard value" << endl <<
	  "for maximum impact parameter (400 m)" <<endl;


	if (fTrigNSB[0] > 0)
	  *fLog << endl << all <<
	    "WARNING: NSB rate can be overestimated by up to 5%." << endl <<
	    "For a precise estimate of the total rate including NSB" << endl << 
	    "accidental triggers I need a file containing all event headers." 
		<< endl;
	else 
	  *fLog << endl << all <<
	    "WARNING: calculating only shower rate (no NSB accidental triggers)" << endl;
      }

    *fLog << endl << all <<
      "WARNING: I will assume the standard maximum off axis angle" << endl <<
      "(5 degrees) for the original showers" << endl;

    for (UInt_t i=0; i<fNum; i++)
      {
	if (fMcRunHeader->GetAllEvtsTriggered())
	  {
	    GetRate(i)->SetImpactMin(0.);
	    GetRate(i)->SetImpactMax(40000.);   // in cm
	  }
	GetRate(i)->SetSolidAngle(2.390941702e-2);  // sr

	//
	// Set limits of energy range, coverting from GeV to TeV:
	//
	GetRate(i)->SetEnergyMin(fMcCorRunHeader->GetELowLim()/1000.);
	GetRate(i)->SetEnergyMax(fMcCorRunHeader->GetEUppLim()/1000.);

      }

    return kTRUE;
}

// --------------------------------------------------------------------------
//
//  overloaded constructor I
//
//      dim:     fDimension
//      *trigbg: number of NSB triggers for
//               a given trigger condition.
//      simbg:   Number of simulated events for the NSB
//      rate:    rate of incident showers
//

MMcTriggerRateCalc::MMcTriggerRateCalc(float rate, int dim,
                                       float *trigbg, float simbg,
                                       const char *name, const char *title)
{
    Init(dim, trigbg, simbg, name, title);
}


// --------------------------------------------------------------------------
//
//  overloaded constructor II
//
//      dim:     fDimension
//      *trigbg: number of NSB triggers for
//               a given trigger condition.
//      simbg:   Number of simulated events for the NSB
//
MMcTriggerRateCalc::MMcTriggerRateCalc(int dim, float *trigbg,float simbg,
                                       const char *name, const char *title)
{
    Init(dim, trigbg, simbg, name, title);
}

MMcTriggerRateCalc::~MMcTriggerRateCalc()
{
    if (fMcTrig)
        delete fMcTrig;

    if (fMcRate)
        delete fMcRate;
}


// --------------------------------------------------------------------------
//
//  The PreProcess connects the raw data with this task. It checks if the 
//  input containers exist, if not a kFalse flag is returned. It also checks
//  if the output contaniers exist, if not they are created.
//  This task can read either Montecarlo files with multiple trigger
//  options, either Montecarlo files with a single trigger option.
//
Bool_t MMcTriggerRateCalc::PreProcess (MParList *pList)
{
    // connect the raw data with this task

    fMcEvt = (MMcEvt*)pList->FindObject("MMcEvt");
    if (!fMcEvt)
    {
        *fLog << err << dbginf << "MMcEvt not found... aborting." << endl;
        return kFALSE;
    }

    UInt_t num;

    fMcTrig = new TObjArray(pList->FindObjectList("MMcTrig", fFirst, fLast));
    num = fMcTrig->GetEntriesFast();
    if (num != fNum)
    {
        *fLog << err << dbginf << fNum << " MMcTrig objects requested, ";
        *fLog << num << " are available... aborting." << endl;
        return kFALSE;
    }

    fMcRate = new TObjArray(pList->FindObjectList("MHMcRate", fFirst, fLast));
    num = fMcRate->GetEntriesFast();
    if (num != fNum)
    {
        *fLog << err << dbginf << fNum << " MHMcRate objects requested, ";
        *fLog << num << " are available... aborting." << endl;
        return kFALSE;
    }

    for (UInt_t i=0;i<fNum;i++)
      {
        MHMcRate &rate = *GetRate(i);

	if (fTrigNSB)
	  rate.SetBackground(fTrigNSB[i], fSimNSB);
	else
	  rate.SetBackground(0., fSimNSB);
      }

    return kTRUE;
}

// --------------------------------------------------------------------------
//
//  The Process-function counts the number of simulated showers, the
//  number of analised showers and the number of triggers. It also updates
//  the limits for theta, phi, energy and impact parameter in the
//  MHMcRate container.
//
Bool_t MMcTriggerRateCalc::Process()
{
    //
    //  Counting analysed showers (except in the case in which the file 
    //  contains only events that triggered, since then we do not know
    //  how many showers were analysed).
    //

    if ( ! fMcRunHeader->GetAllEvtsTriggered() &&
	 fMcEvt->GetPhotElfromShower() )
      fAnalShow++;

    //
    // Set PartId and check it is the same for all events
    //
    if (fPartId < 0)
      fPartId = fMcEvt->GetPartId();
    else if (fPartId != fMcEvt->GetPartId())
      {
	*fLog << err << dbginf << "Incident particle type seems to change";
	*fLog << "from " << fPartId << " to " << fMcEvt->GetPartId() << endl;
	*fLog << "in input data files... aborting." << endl;
	return kFALSE;
      }

    //
    //  Getting angles, energy and impact parameter to set boundaries
    //
    const Float_t theta = fMcEvt->GetTheta();
    const Float_t phi   = fMcEvt->GetPhi();
    const Float_t param = fMcEvt->GetImpact();
    const Float_t ener  = fMcEvt->GetEnergy()/1000.0;

    //
    //  Counting number of triggers
    //
    for (UInt_t i=0; i<fNum; i++)
    {
	if (GetTrig(i)->GetFirstLevel())
	    fTrigger[i] ++; 

	if ( ! fMcRunHeader->GetAllEvtsTriggered())
	  GetRate(i)->UpdateBoundaries(ener, theta, phi, param);
    }

    return kTRUE;
}

// --------------------------------------------------------------------------
//
//  The PostProcess-function calculates and shows the trigger rate
//
Bool_t MMcTriggerRateCalc::PostProcess()
{
    for (UInt_t i=0; i<fNum; i++)
    {
        MHMcRate &rate = *GetRate(i);

        rate.SetParticle(fPartId);
        switch (fPartId)
        {
        case kPROTON:
	  if (fMcCorRunHeader->GetSlopeSpec() != -2.75)
	    {
	      *fLog << err << dbginf <<
		"Spectrum slope as read from  input file ("<< 
		fMcCorRunHeader->GetSlopeSpec() << ") does not match the expected one (-2.75) for protons" << endl << "... aborting." << endl;
	      return kFALSE;
	    }
	  rate.SetFlux(0.1091, 2.75);
	  break;
        case kHELIUM:
	  if (fMcCorRunHeader->GetSlopeSpec() != -2.62)
	    {
	      *fLog << err << dbginf <<
		"Spectrum slope as read from  input file ("<< 
		fMcCorRunHeader->GetSlopeSpec() << ") does not match the expected one (-2.75) for Helium" << endl << "... aborting." << endl;
	      return kFALSE;
	    }
	  rate.SetFlux(0.0660, 2.62);
	  break;
        default:
	  *fLog << err << dbginf << "Unknown incident flux parameters for ";
	  *fLog << fPartId<< " particle Id ... aborting." << endl;
	  return kFALSE;
	}
    }

    //
    // Computing trigger rate
    //
    for (UInt_t i=0; i<fNum; i++)
        GetRate(i)->CalcRate(fTrigger[i], fAnalShow, fShowers);

    return kTRUE;
}
